// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
{**
   Implementation of Transport and TransportServer.
 }
unit transp;

interface

{$I dorb.inc}

uses
  orbtypes,orb,Classes,addr_int,disp_int,{$IFDEF WIN32}winsock{$ENDIF} {$IFDEF LINUX}Libc{$ENDIF},
  code_int,orb_int,env_int,mtmanager_int,mtmanager;

{$IFNDEF USELIB}

type

  TPrincipal = class(TInterfacedObject,IPrincipal)
  private
    FTransport: ITransport;
    FInfo : OctetSeq;
  protected
    procedure encode(const enc: IEncoder);
    function decode(const dec: IDecoder): Boolean;
    function get_property(const prop_name: string): IAny; virtual;
  public
    constructor Create(transp: ITransport = nil);
    class procedure set_peer_info(info: OctetSeq);
  end;

  TTCPTransport = class(TInterfacedObject,ITransport,IORBDispCallBack)
  protected
    FSocket : TSocket;
    FIsBlocking : Boolean;
    FEOF : Boolean;
    FIsBuffering : Boolean;
    FErrors : TList;
    FReadDispatcher,FWriteDispatcher : IORBDispatcher;
    FReadCallBack,FWriteCallback : ITransportCallBack;
    FPeerAddress : IAddress;
    FState: TTransportState;
  public
    procedure block(val: Boolean);
    function connect(const addr : IAddress): Boolean; stdcall;
    procedure open(fd: Longint); stdcall;
    procedure close(); stdcall;
    function write(const buffer;size: _ulong): integer; stdcall;
    function eof(): Boolean; stdcall;
    function bad(): Boolean; stdcall;
    function read(var buffer;size: _ulong): integer; stdcall;
    procedure callback(const disp: IORBDispatcher;event: TDispEvent);
    procedure rselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
    procedure wselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
    function bufread(buf: IBuffer;size: _ulong): integer; stdcall;
    function peer(): IAddress;
    function get_principal(): IPrincipal;
  public
    constructor CreateSock(s : TSocket);
    constructor Create();
    destructor Destroy(); override;
  end;

{$IFDEF HAVE_THREADS}
  TTCPTransportServer = class;

  ITransportServerOperation = interface(IOperation)
  ['{CD4D63CB-8E4F-4CA1-A2AB-353349AB8628}']
    procedure set_server_callback(const AServCallback: ITransportServerCallback);
  end;

  TTransportServerOperationImpl = class(TOperation, ITransportServerOperation{IOperation})
  private
    FCb : ITransportServerCallback;
    FTranServ: ITransportServer;
  protected
    procedure init_shutdown(); override;
    procedure run(); override;
    function copy(): IOperation; override;
    procedure set_server_callback(const AServCallback: ITransportServerCallback);
  public
    constructor Create(const ATranServ: ITransportServer);
  end;
{$ENDIF}

  TTCPTransportServer = class(TInterfacedObject,ITransportServer,{$IFDEF HAVE_THREADS}IOperation,{$ENDIF}IORBDispCallBack)
  private
    FLocalAddress : IAddress;
    FSocket : TSocket;
    FIsBlocking : Boolean;
    FListening : Boolean;
    FDisp : IORBDispatcher;
    FCb : ITransportServerCallback;
    //FOperationObj: TTransportServerOperationImpl;
    FOperation: ITransportServerOperation;//IOperation;
{$IFDEF LINUX}
    FClosed: boolean;
{$ENDIF}
  protected
    function accept(): ITransport;
    procedure aselect(const disp: IORBDispatcher;const cb: ITransportServerCallback);
    function get_address: IAddress;
    function bind(const addr: IAddress): Boolean;
    procedure callback(const disp: IORBDispatcher;event: TDispEvent);
    procedure listen();
    procedure block(doblock: boolean);
    procedure close();
{$IFDEF HAVE_THREADS}
    procedure create_thread();
    procedure start();
    property Operation: ITransportServerOperation{IOperation} read FOperation implements IOperation;
    //property Operation: TTransportServerOperationImpl read FOperationObj implements IOperation;
{$ENDIF}
  public
    constructor Create();
  end;

procedure Principal_set_peer_info(info: OctetSeq);

implementation

uses
  SysUtils, address, any, osnet, throw, except_int;

{$IFDEF WIN32}
const
  SHUT_RDWR = 2; { No more receptions or transmissions. }
{$ENDIF}

threadvar
  _peer_info: OctetSeq;

//********************************************************************
// TTCPTransport
//********************************************************************
constructor TTCPTransport.Create();
begin
  inherited Create();
  FErrors := TList.Create;
  FPeerAddress := TInetAddress.Create('0.0.0.0',0,STREAM);
  FState := tsClosed;
end;

constructor TTCPTransport.CreateSock(s : TSocket);
begin
  Create();
  FSocket := s;
  if FSocket < 0 then
    FSocket := socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (FSocket = -1) then
    dorb_throw(st_COMM_FAILURE, MinorSocket, COMPLETED_NO);
  TOSNet.sockblock(FSocket,true);
  TOSNet.sockndelay(FSocket,true);
  FIsBlocking := True;
  FIsBuffering := False;
  FEof := FALSE;
  //FErrors := TList.Create;
  FState := tsOpen;
end;

destructor TTCPTransport.Destroy();
begin
  if assigned(FReadDispatcher) and Assigned(FReadCallBack) then
    begin
      FReadDispatcher.remove(self,_Read);
      FReadDispatcher := nil;
      FReadCallBack.callback(self,tcRemove);
    end;
  if assigned(FWriteDispatcher) and Assigned(FWriteCallBack) then
    begin
      FWriteDispatcher.remove(self,_Write);
      FWriteDispatcher := nil;
      FWriteCallback.callback(self,tcRemove);
    end;
  TOSNet.sockclose(FSocket);
  //close();
  FErrors.Free;
  inherited Destroy();
end;

procedure TTCPTransport.wselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
begin
  if (FWriteDispatcher <> nil) and (FWriteCallback <> nil) then
    begin
      FWriteDispatcher.remove(self,orbtypes._Write);
      FWriteDispatcher := nil;
      FWriteCallback := nil;
    end;
  if cb <> nil then
    begin
      disp.socketevent(self,FSocket,orbtypes._Write);
      FWriteDispatcher := disp;
      FWriteCallback := cb;
    end;
end;


procedure TTCPTransport.rselect(const disp: IORBDispatcher;const cb: ITransportCallback);
begin
  if (FReadDispatcher <> nil) and (FReadCallback <> nil) then
    begin
      FReadDispatcher.remove(self,orbtypes._Read);
      FReadDispatcher := nil;
      FReadCallback := nil;
    end;
  if cb <> nil then
    begin
      disp.socketevent(self,FSocket,orbtypes._Read);
      FReadDispatcher := disp;
      FReadCallback := cb;
    end;
end;

function TTCPTransport.peer(): IAddress;
var
  sa : TSockAddrIn;
  res: integer;
{$IFDEF WIN32}
  r: integer;
{$ENDIF}
{$IFDEF LINUX}
  r: cardinal;
{$ENDIF}
begin
  result := nil;
  r := sizeof(sa);
  res := getpeername(FSocket,sa,r);
  if res < 0 then
    TOSNet.seterrno()
    //allow for peer() after disconnect ...
    //exit;
  else
    FPeerAddress.set_socketaddress(sa);
  result := FPeerAddress;
end;

procedure TTCPTransport.block(val: Boolean);
begin
  if FIsBlocking <> val then
    begin
      FIsBlocking := val;
      TOSNet.sockblock(FSocket,val);
    end;
end;

procedure TTCPTransport.close();
begin
  if FSocket >= 0 then begin
    {$IFDEF WIN32}WinSock{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}.shutdown(FSocket, SHUT_RDWR);
    TOSNet.sockclose(FSocket);
  end;
  {FSocket := socket(PF_INET,SOCK_STREAM, 0);
  assert(FSocket >= 0);
  TOSNet.sockndelay(FSocket, true);}
  FSocket := -1;
  if assigned(FReadDispatcher) and Assigned(FReadCallBack) then
    FReadDispatcher.remove(self,_Read);
  if assigned(FWriteDispatcher) and Assigned(FWriteCallBack) then
    FWriteDispatcher.remove(self,_Write);
  FIsBuffering := false;
  FIsBlocking := true;
  FReadDispatcher := nil;
  FWriteDispatcher := nil;
  FReadCallBack := nil;
  FWriteCallback := nil;
  {if FSocket >= 0 then
    sockclose(FSocket);}
  FErrors.Clear;
  FEOF := false;
  FState := tsClosed;
end;

procedure TTCPTransport.open(fd: Integer);
{$IFDEF MSWINDOWS}
var
  bufsz: integer;
{$ENDIF}
begin
  Assert(FState = tsClosed);
  if fd < 0 then
    fd := socket(PF_INET,SOCK_STREAM, IPPROTO_TCP);
  if (fd = -1) then
    dorb_throw(st_COMM_FAILURE, MinorSocket, COMPLETED_NO);
  if (TOSNet.sockndelay(fd,true) = -1) then begin
    TOSNet.sockclose(fd);
    dorb_throw(st_COMM_FAILURE, MinorSetsockopt, COMPLETED_NO);
  end;
{$IFDEF MSWINDOWS}
  // Set a larger send buffer size to avoid a performance problem
  // on Win32.
  //
  // TODO: This is a total hack!
  // Thanks guys from OOC
  bufsz := 64 * 1024; // 64K
  if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, PChar(@bufsz), sizeof(integer)) = -1) then begin
    TOSNet.sockclose(fd);
    dorb_throw(st_COMM_FAILURE, MinorSetsockopt, COMPLETED_NO);
  end;
{$ENDIF}
  FSocket := fd;
  FIsBlocking := True;
  FIsBuffering := False;
  FEof := FALSE;
  FState := tsOpen;
end;

function TTCPTransport.connect(const addr : IAddress): Boolean;
var
  sa : TSockAddrIn;
  r : longint;
begin
  sa := addr.socketaddress();
  r := {$IFDEF WIN32}WinSock{$ENDIF} {$IFDEF LINUX}Libc{$ENDIF}.connect(FSocket,sa, sizeof (sa));
  result :=  r >= 0;
end;

function TTCPTransport.write(const buffer;size: _ulong): integer;
var
  written : longint;
  temp : Pointer;
begin
  result := 0;
  temp := @buffer;
  while size > 0 do
    begin
      written := TOSNet.sockwrite(FSocket,temp^,size);
      if written <= 0 then
        break
      else
        begin
          inc(result,written);
          temp := Pointer(Longint(@buffer)+result);
          dec(size,written);
        end;
    end;
end;

function TTCPTransport.eof(): Boolean;
begin
  result := FEOF;
end;

function TTCPTransport.bad(): Boolean;
begin
  result := FErrors.Count > 0;
end;

procedure TTCPTransport.callback(const disp: IORBDispatcher;event: TDispEvent);
begin
  case event of
    orbtypes._Read : FReadCallback.callback(self,orbtypes.tcRead);
    orbtypes._Write : FWriteCallback.callback(self,orbtypes.tcWrite);
    orbtypes._Remove : begin
        FReadCallback := nil;
        FWriteCallback := nil;
        FReadDispatcher := nil;
        FWriteDispatcher := nil;
      end;
    orbtypes._Moved : begin
        FWriteDispatcher := disp;
        FReadDispatcher := disp;
      end;
    else
      assert(false,'not implemented');
  end;
end;

function TTCPTransport.read(var buffer;size: _ulong): integer;
var
  r,todo : longint;
  buf : Pointer;
begin
  todo := size;
  buf := @buffer;
  while todo > 0 do
    begin
      r := TOSNet.sockread(FSocket, buf^, todo);
      if r = SOCKET_ERROR then
        begin
          //seterrno();
          if errno() = EINTR then
            continue;
          if (errno() = 0) or
             (errno() = EWOULDBLOCK) or
             (errno() = EAGAIN) or
             (todo <> longint(size)) then
            break;
          result := r;
          exit;
        end
      else if r = 0 then
        begin
          Feof := true;
          break;
        end
      else
        begin
          dec(todo,r);
          buf := Pointer(Integer(buf)+ r);
        end;
    end;
  result := longint(size) - todo;
end;

function TTCPTransport.bufread(buf: IBuffer;size: _ulong): integer;
var
  i : integer;
begin
  buf.resize(size);
  i := read(Pointer(_ulong(buf.buffer)+buf.WPosition)^,size);
  if i > 0 then
    buf.WseekFromCurrent(i);
  result := i;
end;

function TTCPTransport.get_principal: IPrincipal;
begin
  result := TPrincipal.Create(self);
end;

//********************************************************************
// TTCPTransportServer
//********************************************************************
constructor TTCPTransportServer.Create();
begin
  FLocalAddress := TInetAddress.Create('0.0.0.0',0,STREAM);
  FSocket := socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (FSocket = -1) then
    dorb_throw(st_COMM_FAILURE, MinorSocket, COMPLETED_NO);
  TOSNet.sockreuse(FSocket,True);
  FIsBlocking := true;
  FListening := false;
  FDisp := nil;
  FCb := nil;
  //FOperationObj := TTransportServerOperationImpl.Create;
  //FOperationObj.FTranServ := Self;
  FOperation := TTransportServerOperationImpl.Create(Self);//FOperationObj;
end;

function TTCPTransportServer.accept(): ITransport;
var
  sock : TSocket;
  addr: sockaddr_in;
{$IFDEF MSWINDOWS}
  addrlen: integer;
{$ENDIF}
{$IFDEF LINUX}
  addrlen: socklen_t;
{$ENDIF}
begin
  listen();
  addrlen := SizeOf(sockaddr_in);
  sock := {$IFDEF MSWINDOWS}WinSock{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}.accept(FSocket, @addr, @addrlen);
{$IFDEF LINUX}
  if FClosed then Exit;
{$ENDIF}
  if sock >= 0 then begin
    result := TTCPTransport.Create();
    result.open(sock);
  end;
end;

procedure TTCPTransportServer.callback(const disp: IORBDispatcher;event: TDispEvent);
begin
  case event of
    _Read:
       fcb.ts_callback(self,orbtypes.tsAccept);
    _Remove: begin
        FDisp := nil;
        FCb := nil;
      end;
    _Moved: FDisp := disp;
    else
      assert(false,'not implemented');
  end;
end;

procedure TTCPTransportServer.listen();
var
  res : integer;
begin
  if not FListening then
    begin
      res := {$IFDEF WIN32}WinSock{$ENDIF} {$IFDEF LINUX}Libc{$ENDIF}.listen(FSocket,10);
      if res = 0 then FListening := true;
    end;
end;

procedure TTCPTransportServer.block(doblock: boolean);
begin
  if doblock <> FIsBlocking then begin
    FIsBlocking := doblock;
    TOSNet.sockblock(FSocket, doblock);
  end;
end;

procedure TTCPTransportServer.close;
{$IFDEF LINUX}
var
  sa: TSockAddr;
  s: TSocket;
{$ENDIF}
begin
  if FOperation <> nil then
    FOperation.init_shutdown();
{$IFDEF LINUX}
  // Under Linux __close is not interrupt accept, I haven't any idea why.
  // Connect to the FSocket and interrupt accept, set FClosed in true.
  FClosed := true;
  sa := get_address.socketaddress;
  s := socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  connect(s, sa, SizeOf(sa));
  __close(s);
{$ENDIF}
  {$IFDEF WIN32}WinSock{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}.shutdown(FSocket, SHUT_RDWR);
  TOSNet.sockclose(FSocket);
  FSocket := {$IFDEF WIN32}WinSock{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}.socket(PF_INET,SOCK_STREAM, IPPROTO_TCP);
  Assert(FSocket >= 0);
  FIsBlocking := true;
  TOSNet.sockreuse(FSocket, true);
  FListening := false;
  if FOperation <> nil then
    FOperation.finalize_shutdown();
end;

procedure TTCPTransportServer.aselect(const disp: IORBDispatcher;
  const cb: ITransportServerCallback);
{var
  op: IOperation;}
begin
  if (FDisp <> nil) and (FCb <> nil) then
    begin
      FDisp.remove(self,orbtypes._Read);
      FDisp := nil;
      FCb := nil;
      //FOperationObj.FCb := nil;
      //FOperationObj.FTranServ := nil;
      //-------
      //op := Self as IOperation;
      //FOperation.init_shutdown();
      //sockclose(FSocket);
      //close();
      //FOperation.finalize_shutdown();
      //-------
    end;
  if cb <> nil then
    begin
      listen();
      disp.socketevent(self,FSocket,orbtypes._Read);
      FDisp := disp;
      FCb := cb;
      //FOperationObj.FCb := FCb;
      FOperation.set_server_callback(FCb);
    end;
end;

function TTCPTransportServer.bind(const addr: IAddress): Boolean;
var
  sa : TSockAddrIn;
  r : longint;
begin
  sa := addr.socketaddress();
  r := {$IFDEF WIN32}WinSock{$ENDIF} {$IFDEF LINUX}Libc{$ENDIF}.bind(FSocket,sa, sizeof (sa));
  result :=  r >= 0;
  if not result then
    dorb_throw(st_COMM_FAILURE, MinorBind, COMPLETED_NO);
end;

function TTCPTransportServer.get_address: IAddress;
var
  sa : TSockAddrIn;
  r : longint;
{$IFDEF WIN32}
  len: integer;
{$ENDIF}
{$IFDEF LINUX}
  len: cardinal;
{$ENDIF}
begin
  result := nil;
  len := SizeOf(sa);
  r := {$IFDEF WIN32}WinSock{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}.getsockname(FSocket,sa,len);
  if r < 0 then exit;
  FLocalAddress.set_socketaddress(sa);
  result := FLocalAddress;
end;

procedure TTCPTransportServer.create_thread;
var
  kt: IWorkerThread;
begin
  // create a new thread
  // there is no limit on how much TransportServer we can have
  // each on runs in its own thread !!
  kt := TPM().get_thread_pool(odAccept).get_idle_thread();
  kt.register_operation(FOperation{Self} as IOperation);
  kt.mark_busy();
end;

procedure TTCPTransportServer.start;
{var
  op: IOperation;}
begin
  //op := Self as IOperation;//FOperationObj;
  FOperation.get_thread().post_state_change();
  //FOperation.thread.post_state_change();
end;

//*********************************************************************
//  TPrincipal
//*********************************************************************
constructor TPrincipal.Create(transp: ITransport);
begin
  FInfo := Copy(_peer_info);
  FTransport := transp;
end;

class procedure TPrincipal.set_peer_info(info: OctetSeq);
begin
  _peer_info := copy(info);
end;

function TPrincipal.decode(const dec: IDecoder): Boolean;
var
  u : _ulong;
begin
  result := false;
  if not dec.seq_begin(u) then exit;
  setLength(FInfo,u);
  if u > 0 then
    begin
      if dec.buffer.length < u then exit;
      dec.get_octets(Pointer(FInfo)^,u);
    end;
  if not dec.seq_end then exit;
  result := true;
end;

procedure TPrincipal.encode(const enc: IEncoder);
begin
  enc.seq_begin(Length(FInfo));
  if Length(FInfo) > 0 then
    enc.put_octets(Pointer(FInfo)^,Length(FInfo));
  enc.seq_end;
end;

function TPrincipal.get_property(const prop_name: string): IAny;
begin
  result := CreateAny();
  if CompareStr(prop_name, 'peer-info') = 0 then begin
    //result.set_
  end
  else if CompareStr(prop_name, 'auth-method') = 0 then
    result.put_string('basic')
  else if CompareStr(prop_name, 'peer-address') = 0 then
    result.put_string(FTransport.peer.toString);
end;

procedure Principal_set_peer_info(info: OctetSeq);
begin
  TPrincipal.set_peer_info(info)
end;

{ TTransportServerOperationImpl }

procedure TTransportServerOperationImpl.run;
begin
  while FCb <> nil do
    FCb.ts_callback(FTranServ, tsAccept);
end;

constructor TTransportServerOperationImpl.Create(const ATranServ: ITransportServer);
begin
  inherited Create();
  FTranServ := ATranServ;
end;

function TTransportServerOperationImpl.copy: IOperation;
begin

end;

procedure TTransportServerOperationImpl.init_shutdown;
begin
  FCb := nil;
  FTranServ := nil;
end;

procedure TTransportServerOperationImpl.set_server_callback(const AServCallback: ITransportServerCallback);
begin
  FCb := AServCallback;
end;

{$IFDEF LIBRARY}
exports
  Principal_set_peer_info;
{$ENDIF}
{$ELSE}
procedure Principal_set_peer_info(info: OctetSeq); external MTDORB_Library_Name;
implementation
{$ENDIF}
end.
